#include <iostream>
#include "Observer.h"

using namespace std;
using namespace patterns;

// So you can change POLICY to Unmonitored easily and watch the expected runtime crash.
#define POLICY Monitored 

//
// The Message Types
//
struct FoodMsg {
	int nutrition;
	FoodMsg(int n) : nutrition(n) {}; // Allow type conversion
};

struct SocialMsg {
	int comfort;
	SocialMsg(int c) : comfort(c) {}; // Allow type conversion
};

//
// The Observer
//
class Sim : 
	public Observer<FoodMsg, POLICY>, 
	public Observer<SocialMsg, POLICY>
{
public:
	explicit Sim(const char *s) : name_(s) {};
	void Update(FoodMsg m) { cout << name_.c_str() << " Update(FoodMsg) = " << m.nutrition << endl; };
	void Update(SocialMsg m) { cout << name_.c_str() << " Update(SocialMsg) = " << m.comfort << endl; };
private:
	std::string name_;
};

//
// The Subjects
//
class Oven :
	public Subject<FoodMsg, POLICY>
{
};

class Hottub :
	public Subject<FoodMsg, POLICY>,	// There's a wetbar
	public Subject<SocialMsg, POLICY>	// And attractive members of the preferred gender.
{
public:
	void DoFoodNotify() { patterns::Notify(this, FoodMsg(9)); };		// Resolve by cast
	void DoComfortNotify() { patterns::Notify<SocialMsg>(this, 0); };	// Resolve by specification
};


main()
{
	Sim *arline = new Sim("Arline");
	Sim *richard = new Sim("Ricard");
	Oven *oven = new Oven();

	// When there's no method resolution ambiguity, things are easy!
	Attach(oven, arline);
	oven->Attach(richard);
	
	// Three ways to invoke Subject::Notify() when there's no ambiguity
	oven->Notify(1);			    // traditional
	Notify<FoodMsg>(oven, 2);	    // explicit specification
	Notify(oven, FoodMsg(3));	    // let the compiler resolve

	Hottub *hottub = new Hottub();
	// But watch out for overload resolution that is ambiguous to the programmer.
	Attach(hottub, arline);         // This is bad:  Which Attach() occurred, Food or Social?
		
	// When there is ambiguity, then the helper function is easier to read.
	Attach<SocialMsg>(hottub, arline);                // Policy inferred!
	hottub->Subject<FoodMsg, POLICY>::Attach(arline); // Bad:  Have to name the policy if not default.

	// Three ways to invoke Subject::Notify and resolve ambiguity
	hottub->Subject<SocialMsg, POLICY>::Notify(100);  // Bad:  Non-default policy must be named.
	Notify<SocialMsg>(hottub, 200);
	Notify<FoodMsg>(hottub, 4);
	
	delete richard;					// oven better be notified that richard is deleted...
	Notify(oven, FoodMsg(5));       // ...will cause crash with LifePolicy Unmanaged					
	delete oven;
	delete arline;
	delete hottub;

	return(0);
}


